@@ -64,80 +64,90 @@ def single_source_shortest_paths(graph, s, d=None, annex=None, cost_func=None,
 
     """
     # Current known costs of paths from s to all nodes that have been
-    # reached so far
+    # reached so far. Note that "reached" is not the same as "visited".
     costs = {s: 0}
 
-    # Used with heapq to maintain a priority queue of nodes with known
-    # costs from s
-    open = [(0, s)]
+    # Predecessor of each node that has been reached from s.
+    predecessors = {s: None}
 
-    # Predecessor of each node that has shortest path from s
-    predecessors = {}
+    # A priority queue of nodes with known costs from s. The nodes in
+    # this queue are candidates for visitation. Nodes are added to this
+    # queue when they are reached (but only if they have not already
+    # been visited).
+    visit_queue = [(0, s)]
 
-    while open:
+    # Nodes that have been visited. Once a node has been visited, it
+    # won't be visited again. Note that in this context "visited" means
+    # a node has been selected as having the lowest known cost (and it
+    # must have been "reached" to be selected).
+    visited = set()
+
+    while visit_queue:
         # In the nodes remaining in the graph that have a known cost
         # from s, find the node, u, that currently has the shortest path
         # from s.
-        cost_of_s_to_u, u = heappop(open)
+        cost_of_s_to_u, u = heappop(visit_queue)
 
-        # The edge crossed to get to u
-        prev_e = predecessors.get(u, None)
+        if u == d:
+            break
+
+        if u in visited:
+            # This will happen when u has been reached from multiple
+            # nodes before being visited (because multiple entries for
+            # u will have been added to the visit queue).
+            continue
+
+        visited.add(u)
 
-        # Get nodes adjacent to u...
         if annex and u in annex:
             neighbors = annex[u]
         else:
-            try:
-                neighbors = graph[u]
-            except KeyError:
-                # u has no outgoing edges
+            neighbors = graph[u] if u in graph else None
+
+        if not neighbors:
+            # u has no outgoing edges
+            continue
+
+        # The edge crossed to get to u
+        prev_e = predecessors[u]
+
+        # Check each of u's neighboring nodes to see if we can update
+        # its cost by reaching it from u.
+        for v in neighbors:
+            # Don't backtrack to nodes that have already been visited.
+            if v in visited:
                 continue
 
-        # ...and explore the edges that connect u to those nodes,
-        # updating the cost of the shortest paths to any or all of
-        # those nodes as necessary. v is the node across the current
-        # edge from u.
-        for v in neighbors:
             e = neighbors[v]
 
-            # Get the cost of the edge running from u to v
-            if cost_func:
-                cost_of_e = cost_func(u, v, e, prev_e)
-            else:
-                cost_of_e = e
+            # Get the cost of the edge running from u to v.
+            cost_of_e = cost_func(u, v, e, prev_e) if cost_func else e
 
             # Cost of s to u plus the cost of u to v across e--this
             # is *a* cost from s to v that may or may not be less than
-            # the current known cost to v
+            # the current known cost to v.
             cost_of_s_to_u_plus_cost_of_e = cost_of_s_to_u + cost_of_e
 
-            # When there is a heuristic function, we use
-            # a "guess-timated" cost, which is the normal cost plus
-            # some other heuristic cost from v to d that is calculated
-            # so as to keep us moving in the right direction (generally
-            # more toward the goal instead of away from it).
+            # When there is a heuristic function, we use a
+            # "guess-timated" cost, which is the normal cost plus some
+            # other heuristic cost from v to d that is calculated so as
+            # to keep us moving in the right direction (generally more
+            # toward the goal instead of away from it).
             if heuristic_func:
                 additional_cost = heuristic_func(u, v, e, prev_e)
                 cost_of_s_to_u_plus_cost_of_e += additional_cost
 
-            if v in costs:
+            if v not in costs or costs[v] > cost_of_s_to_u_plus_cost_of_e:
                 # If the current known cost from s to v is greater than
                 # the cost of the path that was just found (cost of s to
                 # u plus cost of u to v across e), update v's cost in
                 # the cost list and update v's predecessor in the
-                # predecessor list (it's now u)
-                if costs[v] > cost_of_s_to_u_plus_cost_of_e:
-                    costs[v] = cost_of_s_to_u_plus_cost_of_e
-                    # u is v's predecessor node. e is the ID of the edge
-                    # running from u to v on the shortest known path
-                    # from s to v. We include the edge's other
-                    # attributes too.
-                    predecessors[v] = (u, e, cost_of_e)
-            else:
-                # No path to v had been found previously.
+                # predecessor list (it's now u). Note that if ``v`` is
+                # not present in the ``costs`` list, it's current cost
+                # is considered to be infinity.
                 costs[v] = cost_of_s_to_u_plus_cost_of_e
                 predecessors[v] = (u, e, cost_of_e)
-                heappush(open, (cost_of_s_to_u_plus_cost_of_e, v))
+                heappush(visit_queue, (cost_of_s_to_u_plus_cost_of_e, v))
 
     if d is not None and d not in costs:
         raise NoPathError('Could not find a path from {0} to {1}'.format(s, d))
@@ -161,17 +171,19 @@ def extract_shortest_path_from_predecessor_list(predecessors, d):
         - The total cost of the path
 
     """
-    nodes = []  # Node IDs on the shortest path from s to d
-    edges = []  # Edge IDs on the shortest path from s to d
-    costs = []  # Costs of the edges on the shortest path from s to d
-    u = d
-    while u in predecessors:
-        v, e, cost = predecessors[u]
+    nodes = [d]  # Nodes on the shortest path from s to d
+    edges = []   # Edges on the shortest path from s to d
+    costs = []   # Costs of the edges on the shortest path from s to d
+    predecessor = predecessors[d]
+    while predecessor is not None:
+        # u is the node from which v was reached, e is the edge
+        # traversed to reach v from u, and cost is the cost of u to
+        # v over e. (Note that v is implicit--it's the previous u).
+        u, e, cost = predecessor
         nodes.append(u)
         edges.append(e)
         costs.append(cost)
-        u = v
-    nodes.append(u)  # Start node
+        predecessor = predecessors[u]
     nodes.reverse()
     edges.reverse()
     costs.reverse()
